;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; make all pritimives as objects

(define wrap-primitive
  (lambda (primitive-val)
    (let ((v (make-vector 1)))
      (vector-set! v 0 primitive-val)
      (make-object 'primitive v))))

(define unwrap-primitive
  (lambda (wrapped-primitive)
    (get-field-value 'val 'primitive wrapped-primitive)))

(define lit-exp-datum
  (around-wrap lit-exp-datum
	       (lambda (original-lit-exp-datum exp)
		 (wrap-primitive (original-lit-exp-datum exp)))))

(define apply-primitive
  (around-wrap apply-primitive
	       (lambda (original-apply-primitive prim wrapped-args)
		 (let ((result
			(original-apply-primitive prim
						  (map unwrap-primitive wrapped-args))))
		   (wrap-primitive result)))))

(define elaborate-class-decls!
  (around-wrap elaborate-class-decls!
	       (lambda (original-elaborate-class-decls! decls)
		 (original-elaborate-class-decls! 
		  (append decls
			  (parse-decls
			   '((class primitive object
				    (field object val)
				    (method void init ((object v))
					    (fset! this val v)
					    0)
				    (method object getval () (fget this val))
				    ))))))))

(define maybe-unwrap
  (lambda (maybe-primitive)
    (if (and (object? maybe-primitive)
	     (eq? (object-cname maybe-primitive) 'primitive))
	(unwrap-primitive maybe-primitive)
	maybe-primitive)))

(define eval-program
  (around-wrap eval-program
   (lambda (original-eval-program pgm)
     (maybe-unwrap (original-eval-program pgm)))))

(define eval-exp
  (around-wrap eval-exp
   (lambda (original-eval-exp exp env)
     (if (if-exp? exp)
	 (if (maybe-unwrap (eval-exp (if-exp-test-exp exp) env))
	     (eval-exp (if-exp-then-exp exp) env)
	     (eval-exp (if-exp-else-exp exp) env))
	 (original-eval-exp exp env)))))

