#|<--                                92 columns                                         ->|#

(let ((time-stamp "Time-stamp: <03/05/20 00:56:24 kazu>"))
  (display (string-append "weaver=padflow-rt.scm: " (substring time-stamp 13 29)))
  (newline))

(load "pureoo.scm")
(load "weaver=paex-rt.scm")

(define-struct (call-jp struct:jp)	;a join point has a slot for a return value,
  (mname target args return))		;which is void when created.

(set! eval-program
  (around-wrap eval-program
    (lambda (original-eval-program pgm)
      (collect-dflow-advice-decls! *advice-decls*)
      (original-eval-program pgm))))
      
(define call-method
  (lambda (mname obj args)
    (let* ((jp                (make-call-jp .stack-previous-jp.
					    mname obj args
				      '*void-return-value*)) ;JPs are method calls in X
	   (method            (lookup-method jp))            ;ID-A
	   (adv-matches       (lookup-advice jp))            ;ID-B
	   (actions-on-result (process-dflow-mark! jp)))     ;(val->unit) list
      (fluid-let ((.stack-previous-jp. jp))
	(let ((result
	       (execute-advice adv-matches jp ;effect-B: run the advice code
		 (lambda (args)
		   (execute-method method			 ;effect-A:
				   (jp-replace-args jp args))))));run the method body

	  (for-each (lambda (action) (action result))            ;process remaning marking
		    actions-on-result)				 ;actions to the returned
								 ;value from the join point
	  result)))))						 

(set! jp-replace-args
      (around-wrap jp-replace-args
       (lambda (jp-replace-args jp args)
	 (if (call-jp? jp)
	     (make-call-jp (jp-stack-prev-jp jp) (call-jp-mname jp) (call-jp-target jp)
			   args (call-jp-return jp))
	     (jp-replace-args jp args)))))

(define process-dflow-mark!;; jp -> (val -> unit) list
  (lambda (jp)
    (let ((process-pcd! ; ptc * param * (val -> unit) -> (val->unit) list
	   (lambda (params pcd param action!)
	     (let ((ptc-match (and (not (eqv? pcd '*void-bypass-pcd*))
				   (pointcut-matches params ; !!!(list param)?
						     pcd jp))))
	       (if ptc-match
		   (let ((val  (ptc-match-lookup ptc-match param jp)))
		     (if (eq? val '*void-return-value*)
			 (list action!)
			 (begin (action! val) '())))
		   '())))))
      (flatmap (lambda (id dflow-pcd params)
		 (append 
		  (process-pcd! params
				(dflow-pcd-source-pcd   dflow-pcd)
				(dflow-pcd-source-param dflow-pcd)
				(add-dflow-mark-object! id))
		  (process-pcd! params
				(dflow-pcd-bypass-pcd      dflow-pcd)
				(dflow-pcd-bypass-param    dflow-pcd)
				(remove-dflow-mark-object! id))))
	       (map dflow-pcd-decl-id     *dflow-pcd-decls*)
	       (map dflow-pcd-decl-pcd    *dflow-pcd-decls*)
	       (map dflow-pcd-decl-params *dflow-pcd-decls*)))))

(define ptc-match-lookup
  (lambda (ptc-match param jp)
    (let* ((args (and (args-jp? jp) (jp-args jp)))
	   (ids  (ptc-match-ids ptc-match))
	   (vals ((ptc-match-args->vals ptc-match) args)))
      (cdr (assq param (map cons ids vals))))))

(define process-dflow-mark-object!; id * (id * (id list) -> (id list)) -> (val -> unit)
  (lambda (dflow-id action)		       ;N.B. an alternative semantics may not
    (lambda (obj)			       ;propagate the action when the action is
      (propagate-dflow-object-fields! obj      ;to remove a tag.!!!
	(lambda (obj)
	  (set-object-dflow! obj (action dflow-id (object-dflow obj))))))))

(define add-dflow-mark-object!; id -> (val -> unit)
  (lambda (dflow-id)
    (process-dflow-mark-object! dflow-id cons)))

(define remove-dflow-mark-object!; id -> (val -> unit)
  (lambda (dflow-id)
    (process-dflow-mark-object! dflow-id remove*)))

(define propagate-dflow-object-fields!
  (lambda (obj action!)
    (let loop ((obj obj) (visited '()))
      (if (and (not (memv obj visited)) (object? obj))
	  (begin
	    (action! obj)
	    (for-each (lambda (fname)
			(loop (get-field fname obj)
			      (cons obj visited)))
		      (object->fnames obj)))))))

(define get-field
  (lambda (fname object)
    (get-field-value fname (object-cname object) object)))

(define object->fnames			
  (lambda (object)
    (cname->fnames (object-cname object))))

(define propagate-dflow-marks!
  (lambda (from-val to-val)
    (if (and (object? from-val) (object? to-val))
	(set-object-dflow! to-val
			   (append (object-dflow from-val)
				   (object-dflow to-val))))))

(define *dflow-pcd-decls* '())

(define-struct dflow-pcd-decl
  (id pcd params))

(define collect-dflow-advice-decls!
  (lambda (advice-decls)
    (set! *dflow-pcd-decls* '())
    (map (lambda (advice)
	   (walk-pcd (advice-decl-pcd advice)
		     (lambda (pcd)
		       (if (dflow-pcd? pcd)
			   (set! *dflow-pcd-decls*
				 (cons (make-dflow-pcd-decl
					(gensym) pcd (advice-decl-params advice))
				       *dflow-pcd-decls*)))
		       pcd)))
	 advice-decls)))

(define dflow-pcd-id
  (lambda (dflow-pcd)
    (dflow-pcd-decl-id
     (find-if (lambda (dflow-pcd-decl)
		(eq? dflow-pcd (dflow-pcd-decl-pcd dflow-pcd-decl)))
	      *dflow-pcd-decls*))))


(define-struct object (cname vals dflow)) ; patch to existing declaration

(set! make-object
  (around-wrap make-object
    (lambda (orinigal-make-object cname vals)
      (orinigal-make-object cname vals '()))))

(set! apply-primitive
  (around-wrap apply-primitive
	       (lambda (original-apply-primitive prim args)
		 (let ((result
			(original-apply-primitive prim args)))
		   (for-each (lambda (arg) (propagate-dflow-marks! arg result))
			     args)
		   result))))

;; dlow's propagation for object 
;(define apply-...)

(set! pointcut-matches
  (around-wrap pointcut-matches
    (lambda (original-pointcut-matches params ptc jp)
      (cond ((dflow-pcd? ptc)
	     (let ((dest-match (pointcut-matches params
						 (dflow-pcd-dest-pcd ptc) jp)))
	       (if dest-match
		   (let ((val (ptc-match-lookup dest-match
						(dflow-pcd-dest-param ptc)
						jp)))
		     (if (memq (dflow-pcd-id ptc)
			       (object-dflow val)) ;(collect-dflow-from-reachable val)?!!!
			 dest-match
			 #f))
		   #f)))
	    ((return-pcd? ptc)
	     (make-ptc-match (list (return-pcd-id  ptc))
			     (lambda (args)
			       (list (call-jp-return jp)))
			     id-args))
	    (else 
	     (original-pointcut-matches params ptc jp))))))
