#|<--                                92 columns                                         ->|#


(let ((time-stamp "Time-stamp: <2002-09-17 11:59:57 Hidehiko Masuhara>"))
  (display (string-append "weaver=paex-rt.scm: " (substring time-stamp 13 29)))
  (newline))

(load "weaver=pa-rt.scm")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; adding more kinds of join ponts

;;; all kinds of jps are now subtypes of jp

(define-struct jp                  ())
(define-struct (call-jp struct:jp) (mname target args))

;;; get-jp for field gets

(define-struct (get-jp  struct:jp) (fname target))

;;; generate a JP for getting a field value

(define  get-field-value
  (lambda (fname within-method-cname obj)
    (let* ((jp (make-get-jp fname obj))
	   (adv-matches (lookup-advice jp)))
      (execute-advice adv-matches jp
	(lambda ()
	  (deref (lookup-field-ref fname within-method-cname obj)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; adding more kind of pointcuts
;;;
;;; 1. for each pointcut, test if the join point is a kind of
;;;    matching.  (e.g., call pointcuts matches only call jps)
;;;
;;; 2. add matching rules for new kind of pointcuts similar to the one
;;;    for call pointcuts.
;;;
;;; 3. take care of some pointcuts that can take values out of more
;;;    than one kind of jps (e.g., target pointcut)

(define-struct ptc-match (ids args->vals args*vals->args))

(define args->null-vals ; args->vals
  (lambda (args) '()))
(define id-args ; args*vals->args
  (lambda (args ignore:vals) args))

(define pointcut-matches
  (lambda (params ptc jp)
    (cond ((call-pcd? ptc)
	   (and (call-jp? jp)		; EXT: check jp type
		(eq? (call-pcd-mname ptc)
		     (call-jp-mname jp))
		(make-ptc-match '() args->null-vals id-args)))

	  ((get-pcd? ptc)		; match get-jp with matching fname
	   (and (get-jp? jp)
		(eq? (get-pcd-fname ptc)
		     (get-jp-fname jp))
		(make-ptc-match '() args->null-vals id-args)))

	  ((target-pcd? ptc)
	   (and (target-jp? jp)		; EXT: check jp type
		(make-ptc-match (list (target-pcd-id ptc))
				(lambda (ignore:args)
				  (list (jp-target jp)))
				id-args))) 
	  
	  ;; given (around ((int a) (int b)) (args b c) (proceed 1 2)) and
	  ;; a call-jp{args=(3 4)}, 
	  ;; then args=(3 4), vals=(1 2), (params->pnames params)=(a b),
	  ;; (args-pcd-ids ptc)=(b c),
	  ;; therefore args*vals->args shall return (2 4)

	  ((args-pcd? ptc)
	   (and (args-jp? jp)		; EXT: check jp type
		(make-ptc-match (args-pcd-ids ptc)
				(lambda (args) args) ; args returns proceeded values
				(lambda (args vals)
				  (let ((env (extend-env (params->pnames params)
							 vals
							 (make-empty-env))))
				    (map (lambda (id arg)
					   (let ((ref (apply-env-ref env id)))
					     (if ref (deref ref) arg)))
					 (args-pcd-ids ptc)
					 args))))))	
	  
	  ((and-pcd? ptc)
	   (let loop ((ptcs       (and-pcd-pcds ptc))
		      (prev-match (make-ptc-match '() args->null-vals id-args)))
	     (if (null? ptcs)
		 prev-match
		 (let ((this-match (pointcut-matches params (car ptcs) jp)))
		   (and this-match
			(loop (cdr ptcs)
			      (ptc-match-append prev-match this-match)))))))

	  ((or-pcd? ptc)
	   (let loop ((ptcs (or-pcd-pcds ptc)))
	     (and (not (null? ptcs))
		  (or (pointcut-matches params (car ptcs) jp)
		      (loop (cdr ptcs))))))
	  ((not-pcd? ptc)
	   (let ((match (pointcut-matches params (not-pcd-pcd ptc) jp)))
	     (if match
		 #f
		 (make-ptc-match '() args->null-vals id-args))))
	  (else (error 'pointcut-matches "inappropriate pointcut ~s" ptc)))))

(define ptc-match-append
  (lambda (m1 m2)
    (make-ptc-match (append (ptc-match-ids  m1) (ptc-match-ids  m2))
		    (lambda (args)
		      (append ((ptc-match-args->vals m1) args)
			      ((ptc-match-args->vals m2) args)))
		    (lambda (args vals)
		      (let* ((args-1 ((ptc-match-args*vals->args m2) args   vals))
			     (args-2 ((ptc-match-args*vals->args m1) args-1 vals)))
			args-2)))))

(define lookup-advice
  (lambda (jp)
    (remove* #f
      (map (lambda (adecl)
	     (let* ((params    (advice-decl-params adecl))
		    (pcd       (advice-decl-pcd adecl))	   ; pcd -> ptc
		    (ptc-match (pointcut-matches params pcd jp))) ; pcd -> ptc
	       (if (not ptc-match)
		   #f
		   (make-adv-match adecl ptc-match))))
	   *advice-decls*))))

;;; the next two predicates discriminates join points that have
;;; appropriate information

(define target-jp? (lambda (jp) (or (call-jp? jp) (get-jp? jp))))
(define args-jp?   (lambda (jp)     (call-jp? jp)))

;;; the next two procedures retrives appropriate information from
;;; different kind of jps

(define jp-target
  (lambda (jp)
    (cond ((call-jp? jp) (call-jp-target jp))
	  ((get-jp?  jp) (get-jp-target  jp))
	  (else		 (error 'jp-target "not appropriate jp ~s" jp)))))

(define jp-args
  (lambda (jP)
    (cond ((call-jp? jp) (call-jp-args jp))
	  (else		 (error 'jp-args "not appropriate jp ~s" jp)))))

(define jp-replace-args
  (lambda (jp args)
    (cond ((call-jp? jp) (make-call-jp (jp-stack-prev-jp jp) (call-jp-mname jp)
				       (call-jp-target jp) args))
	  ((get-jp? jp)  (make-get-jp  (jp-stack-prev-jp jp) (get-jp-fname jp)
				       (get-jp-target jp)));no change
	  (else (error 'jp-replace-args "not appropriate jp ~s" jp)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; after advice
;;;
;;; To add after advice, extend execute-advice so that it runs
;;; after-advices similar to before-advices, but after the thunk is
;;; executed.
;;;
;;; NOTE: since BASE is not capable with exception handling, after
;;; advice in this code really means 'after returning' in AspectJ.

(define execute-advice
  (lambda (adv-matches jp thunk)
    (let ((befores (gather-adv-matches is-before? adv-matches))
	  (afters  (gather-adv-matches is-after?  adv-matches)))
      (for-each (lambda (adv-match)				;run befores
		  (execute-one-advice adv-match jp))
		befores)
      (let ((result (thunk)))					;run thunk
	(for-each (lambda (adv-match)				;run afters
		    (execute-one-advice adv-match jp))
		  afters)
	result))))						;return thunk result

(define gather-adv-matches
  (lambda (pred adv-matches)
    (collect-if (lambda (adv-match)
		  (pred (adv-match-adv-decl adv-match)))
		adv-matches)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; around advice
;;;
;;; To add around advice and proceed mechanism,
;;;
;;; 1. extend execute-advice again so that it first runs all around
;;; advice first, and then runs befores, thunk, and afters.  
;;;
;;; 2. extend execute-one-advice so that it takes a lambda closure
;;; that continues executions of remaining around advice.  The closure
;;; is to be bound to a special variable in the environment of body of
;;; the advice execution.
;;;
;;; 3. extend eval-exp to evaluate proceed expressions to take out the
;;; closure bound in the above process, and call it.

;;;
;;; NOTE: arounds always followed by befores and afters (unlike AspectJ)

(define execute-advice
  (lambda (adv-matches jp thunk)
    (let* ((befores (gather-adv-matches is-before? adv-matches))
	   (afters  (gather-adv-matches is-after?  adv-matches))
	   (arounds (gather-adv-matches is-around? adv-matches)))

      (let loop ((arounds arounds)
		 (args (and (args-jp? jp) (jp-args jp))))	 ;extract args from jp
	(if (not (null? arounds))
	    (execute-one-advice (car arounds) jp args            ;run around
	     (lambda (vals)				         ;thunk is to run next
	       (loop (cdr arounds)				 ;around advice with 
		     (update-args (car arounds) args vals))))    ;updated args
				  		
	    (begin						 ;when no more arounds,
	      (for-each (lambda (adv-match)			 ;run befores
			  (execute-one-advice adv-match jp args #f))
			befores)
	      (let ((result (thunk args)))                       ;run body
		(for-each (lambda (adv-match)			 ;run afters
			    (execute-one-advice adv-match jp args #f))
			  afters)
		result)))))))					 ;return result

(define execute-one-advice
  (lambda (adv-match jp args thunk)				 ; receive a thunk
    (eval-body (advice-decl-body (adv-match-adv-decl adv-match))
	       (make-advice-env adv-match jp args thunk))))	 ; and bound it in env

(define update-args
  (lambda (adv-match args-to-thunk vals-to-proceed)
    (let ((args*vals->args (ptc-match-args*vals->args (adv-match-ptc-match adv-match))))
      (args*vals->args args-to-thunk vals-to-proceed))))

(set! eval-exp
  (around-wrap eval-exp
    (lambda (eval-exp exp env)
      (if (proceed-exp? exp)					 ; evaluate proceed
	  (execute-proceed (deref (apply-env-ref env '%proceed))
			   (eval-rands (proceed-exp-rands exp) env))
	  (eval-exp exp env)))))

(define make-advice-env
  (lambda (adv-match tjp args thunk)
    (let* ((ptc-match (adv-match-ptc-match adv-match))
	   (ids       (ptc-match-ids  ptc-match))
	   (vals      ((ptc-match-args->vals ptc-match) args))) ;args ptc parameters are
								;bound to values of args
								;in the jp, which can be
								;modified by proceed
      (new-env (append '(tjp %proceed) ids)			
	       (append (list tjp thunk) vals)))))

(define execute-proceed
  (lambda (thunk vals-to-proceed)	; vals-to-proceed should be reorderd
    (thunk vals-to-proceed)))		; for the thunk, but not for now

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; cflow advice
;;;
;;; To add cflow pointcut,
;;;
;;; 1. extend the join point strucutre to hold the previous join point
;;;    in terms of calling context.
;;;
;;; 2. extend procedures that creates join points (like call-method)
;;;    so that it properly store the stack-previous join point in the
;;;    created join point.  To do so, we used fluid-let mechanism in
;;;    Scheme.
;;;
;;; 3. extend pointcut-matches to match cflow pointcuts to the join
;;;    points when the body of cflow matches to the join points, or
;;;    the cflow pointcut matches to the stack-previous join points.

;;; jp has a previous jp in terms of calling context
(define-struct jp                  (stack-prev-jp))

;;; the next structure declarations just needs redefinition
(define-struct (call-jp struct:jp) (mname target args))
(define-struct (get-jp  struct:jp) (fname target))

(define .stack-previous-jp. #f)

;;; the next two procedures are rewritten to use make-jp
(define call-method
  (lambda (mname obj args)
    (let* ((jp          (make-call-jp .stack-previous-jp.	;EXT: record sp-jp
				      mname obj args)) 
	   (method      (lookup-method jp))
	   (adv-matches (lookup-advice jp)))
      (fluid-let ((.stack-previous-jp. jp))			;EXT: update sp-jp
	(execute-advice adv-matches jp	   
	  (lambda (args)
	    (execute-method method (jp-replace-args jp args))))))))

(define  get-field-value
  (lambda (fname within-method-cname obj)
    (let* ((jp          (make-get-jp .stack-previous-jp.	;EXT: record sp-jp
				     fname obj))
	   (adv-matches (lookup-advice jp)))
      (fluid-let ((.stack-previous-jp. jp))			;EXT: update sp-jp
	(execute-advice adv-matches jp
	  (lambda (args)
	    (deref (lookup-field-ref fname within-method-cname obj))))))))

;;; (cflow ptc) pointcut matches a jp if ptc matches jp, or 
;;; (cflow ptc) matches a stack previous jp of jp.
(set! pointcut-matches
  (around-wrap pointcut-matches
    (lambda (other-pointcut-matches params ptc jp)
      (letrec ((loop 
		(lambda (ptc jp)
		  (and jp				 ;unmatch if reached bottom
		    (let ((ptc-match 
			   (pointcut-matches params      ;match current jp against
					     ptc         ;body pointcut 
					     jp)))
		      (if ptc-match			 ;match if body matches
			     
			  (make-ptc-match
			   (ptc-match-ids ptc-match)
			   (apply-with-args-in-jp
			    ptc-match jp)
			   id-args)			 ;proceeded args for cflow'd args 
							 ;shall be dropped

			  (loop ptc (jp-stack-prev-jp jp));run down the stack
			  ))))))			  ;if unmatched
								     
	(cond ((cflow-pcd? ptc)
	       (loop (cflow-pcd-pcd ptc) jp))
	      ((cflowbelow-pcd? ptc)
	       (loop (cflowbelow-pcd-pcd ptc)
		     (jp-stack-prev-jp jp)))
	      (else (other-pointcut-matches params ptc jp)))))))

(define apply-with-args-in-jp			 ;args ptc in cflow should take 
  (lambda (ptc-match jp)			 ;args from jps in the stack, not 
    (lambda (ignore:args)			 ;from proceed-modified args
      ((ptc-match-args->vals ptc-match)
       (and (args-jp? jp) (jp-args jp))))))
